"""Arguments"""

from typing import Optional

from .common import Descriptor

ARG_DESCRIPTOR_EMPTY = '#EMPTY#'


class ArgumentDescriptor(Descriptor):
    """A descriptor for arguments"""

    EMPTY = ARG_DESCRIPTOR_EMPTY

    def __init__(
            self,
            name: str,
            type_: Optional[str],
            description: Optional[str],
            default: Optional[str] = ARG_DESCRIPTOR_EMPTY
    ) -> None:
        """A descriptor for arguments

        Args:
            name (str): The argument name
            type_ (Optional[str]): The argument type
            description (Optional[str]): The arguments description
            default (Optional[str], optional): The default value. Defaults to ARG_DESCRIPTOR_EMPTY.
        """
        self.name = name
        self.type = type_
        self.description = description
        self.default = default

    @property
    def is_optional(self) -> bool:
        """Indicates whether the argument is optional

        Note that an argument is optional if it has a default value, not if it
        has the Optional[...] type decoration.

        Returns:
            bool: True if the argument is optional
        """
        return self.default is not ArgumentDescriptor.EMPTY

    @property
    def descriptor_type(self) -> str:
        return "argument"

    def __repr__(self):
        return f'{self.name}: {self.type} - {self.description}'



"""Meta data"""

from .common import Descriptor
from .arguments import ArgumentDescriptor
from .raises import RaisesDescriptor
from .callables import CallableDescriptor, CallableType
from .properties import PropertyDescriptor
from .classes import ClassDescriptor
from .modules import ModuleDescriptor

__all__ = [
    "Descriptor",
    "ArgumentDescriptor",
    "RaisesDescriptor",
    "CallableDescriptor",
    "CallableType",
    "PropertyDescriptor",
    "ClassDescriptor",
    "ModuleDescriptor"
]



"""Meta data"""

from __future__ import annotations
import inspect
from types import ModuleType
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple
)

import docstring_parser

from .arguments import ArgumentDescriptor
from .common import Descriptor
from .classes import ClassDescriptor
from .callables import CallableDescriptor
from .utils import make_file_relative, is_child_module


class ModuleDescriptor(Descriptor):
    """A module descriptor"""

    def __init__(
            self,
            name: str,
            summary: Optional[str],
            description: Optional[str],
            attributes: List[ArgumentDescriptor],
            examples: Optional[List[str]],
            package: Optional[str],
            file: Optional[str],
            classes: List[ClassDescriptor],
            functions: List[CallableDescriptor],
            modules: List[ModuleDescriptor]
    ) -> None:
        """A module descriptor

        Args:
            name (str): The module name
            summary (Optional[str]): The module summary
            description (Optional[str]): The module description
            attributes (List[ArgumentDescriptor]): The attribute list
            examples (Optional[List[str]]): Examples from the docstring
            package (Optional[str]): The package name
            file (Optional[str]): The file name
            classes (List[ClassDescriptor]): Classes in the module
            functions (List[CallableDescriptor]): Functions in the module
            modules (List[ModuleDescriptor]): The child modules
        """
        self.name = name
        self.summary = summary
        self.description = description
        self.attributes = attributes
        self.examples = examples
        self.package = package
        self.file = file
        self.classes = classes
        self.functions = functions
        self.modules = modules

    @property
    def descriptor_type(self) -> str:
        return "module"

    def __repr__(self) -> str:
        return f'{self.name} - {self.summary}'

    @classmethod
    def create(
            cls,
            module: ModuleType,
            class_from_init: bool,
            ignore_dunder: bool,
            ignore_private: bool,
            ignore_all: bool,
            ignore_inherited: bool,
            prefer_docstring: bool,
            follow_module_tree: bool
    ) -> ModuleDescriptor:
        """Create a module descriptor

        Args:
            obj (Any): The module object
            class_from_init (bool): If True take the docstring from the init
                function
            ignore_dunder (bool): If True ignore
                <span>&#95;&#95;</span>XXX<span>&#95;&#95;</span> functions
            ignore_private (bool): If True ignore private methods
                (those prefixed &#95;XXX)
            ignore_all (bool): If True ignore the
                <span>&#95;&#95;</span>all<span>&#95;&#95;</span> member.
            prefer_docstring (bool): If true prefer the docstring
            follow_module_tree (bool): If true follow the module tree

        Returns:
            ModuleDescriptor: A module descriptor
        """
        docstring = docstring_parser.parse(inspect.getdoc(module))

        name = module.__name__
        summary = docstring.short_description if docstring else None
        description = docstring.long_description if docstring else None
        attrs: List[Tuple[str, str]] = [
            (meta.args[1], meta.description)
            for meta in docstring.meta
            if 'attribute' in meta.args
        ]
        attributes: List[ArgumentDescriptor] = []
        for attr_details, attr_desc in attrs:
            attr_name, _sep, attr_type = attr_details.partition(' ')
            attr_type = attr_type.strip('()')
            attributes.append(
                ArgumentDescriptor(attr_name, attr_type, attr_desc)
            )
        examples: Optional[List[str]] = [
            meta.description
            for meta in docstring.meta
            if 'examples' in meta.args
        ] if docstring is not None else None

        package = module.__package__
        file = make_file_relative(module.__file__)

        members: Dict[str, Any] = dict(inspect.getmembers(module))
        valid_members = members.get('__all__', [])

        classes: List[ClassDescriptor] = []
        functions: List[CallableDescriptor] = []
        for member_name, member in members.items():

            if (
                    (not ignore_all and member_name not in valid_members)
                    and inspect.getmodule(member) is not module
            ):
                # Only handler members in this module, or members in __all__ if
                # this is not ignored.
                continue
            if ignore_dunder and member_name.startswith('__') and member_name.endswith('__'):
                continue
            if ignore_private and member_name.startswith('_'):
                continue

            if ignore_all or not valid_members or member_name in valid_members:

                if inspect.isclass(member):
                    classes.append(
                        ClassDescriptor.create(
                            member,
                            class_from_init,
                            ignore_dunder,
                            ignore_private,
                            ignore_inherited,
                            name
                        )
                    )
                elif inspect.isfunction(member):
                    functions.append(
                        CallableDescriptor.create(
                            member,
                            prefer_docstring=prefer_docstring
                        )
                    )
                else:
                    print(f'unknown {member_name}')

        modules = [
            ModuleDescriptor.create(
                member,
                class_from_init,
                ignore_dunder,
                ignore_private,
                ignore_all,
                ignore_inherited,
                prefer_docstring,
                follow_module_tree
            )
            for member in members.values()
            if follow_module_tree and is_child_module(module, member)
        ]

        return ModuleDescriptor(
            name,
            summary,
            description,
            attributes,
            examples,
            package,
            file,
            classes,
            functions,
            modules
        )



"""Callables"""

from __future__ import annotations
from enum import Enum, auto
import inspect
from inspect import Parameter, Signature
from typing import (
    Any,
    List,
    Optional
)

import docstring_parser
from docstring_parser import Docstring

from ..utils import (
    find_docstring_param,
    get_type_name
)

from .arguments import ArgumentDescriptor
from .common import Descriptor
from .raises import RaisesDescriptor
from .utils import make_file_relative


class CallableType(Enum):
    """An enum indicating the type of a callable"""
    FUNCTION = auto()
    CONSTRUCTOR = auto()
    METHOD = auto()
    CLASS_METHOD = auto()


CLASS_FUNCTIONS = {
    CallableType.CLASS_METHOD,
    CallableType.METHOD
}


class CallableDescriptor(Descriptor):
    """A descriptor for a callable"""

    POSITIONAL_ONLY = '/'
    KEYWORD_ONLY = '*'

    def __init__(
            self,
            qualifier: str,
            name: str,
            summary: Optional[str],
            description: Optional[str],
            arguments: List[ArgumentDescriptor],
            return_type: str,
            return_description: Optional[str],
            callable_type: CallableType,
            is_async: bool,
            is_generator: bool,
            raises: Optional[List[RaisesDescriptor]],
            examples: Optional[List[str]],
            module: str,
            package: Optional[str],
            file: Optional[str]
    ) -> None:
        """A descriptor for a callable

        Args:
            qualifier (str): The qualifier part of the name
            name (str): The name of the callable
            summary (Optional[str]): The callables summary docstring
            description (Optional[str]): The callables description docstring
            arguments (List[ArgumentDescriptor]): The callables arguments
            return_type (str): The callables return type
            return_description (Optional[str]): The callables return description
            callable_type (CallableType): The type of callable
            is_async (bool): True if the callable is async
            is_generator (bool): True if the callable is a generator
            raises (Optional[List[RaisesDescriptor]]): A list of the exceptions
                raised
            examples (Optional[List[str]]): A list of examples
            module (str): The module name
            package (Optional[str]): The package name
            file (Optional[str]): The file name
        """
        self.qualifier = qualifier
        self.name = name
        self.summary = summary
        self.description = description
        self.arguments = arguments
        self.return_type = return_type
        self.return_description = return_description
        self.callable_type = callable_type
        self.is_async = is_async
        self.is_generator = is_generator
        self.raises = raises
        self.examples = examples
        self.module = module
        self.package = package
        self.file = file

    @property
    def descriptor_type(self) -> str:
        return "callable"

    def __repr__(self) -> str:
        return f'{self.name} - {self.summary}'

    @property
    def callable_type_description(self) -> str:
        """The function type name.

        One of: 'class', 'method', 'async generator function'
        'generator function', 'function'

        Returns:
            str: The function type name
        """
        if self.callable_type == CallableType.CONSTRUCTOR:
            return 'class'

        description = 'async ' if self.is_async else ''
        description += 'generator ' if self.is_generator else ''

        if self.callable_type == CallableType.METHOD:
            description += 'method'
        elif self.callable_type == CallableType.CLASS_METHOD:
            description += 'class method'
        else:
            description += 'function'

        return description

    @classmethod
    def create(
            cls,
            obj: Any,
            signature: Optional[Signature] = None,
            docstring: Optional[Docstring] = None,
            callable_type: CallableType = CallableType.FUNCTION,
            prefer_docstring=False,
            qualifier: Optional[str] = None
    ) -> CallableDescriptor:
        """Create a callable descriptor from a callable

        Args:
            obj (Any): The callable
            signature (Optional[Signature], optional): The signature. Defaults
                to None.
            docstring (Optional[Docstring], optional): The docstring. Defaults
                to None.
            callable_type (CallableType, optional): The function type. Defaults
                to CallableType.FUNCTION.
            prefer_docstring (bool): If true prefer the docstring.
            qualifier (Optional[str], optional): An overload for the qualifier.
                Defaults to None.

        Returns:
            CallableDescriptor: A callable descriptor
        """
        if signature is None:
            signature = inspect.signature(obj)
        if docstring is None:
            docstring = docstring_parser.parse(inspect.getdoc(obj))

        is_async = inspect.iscoroutinefunction(
            obj) or inspect.isasyncgenfunction(obj)
        is_generator = inspect.isgeneratorfunction(
            obj) or inspect.isasyncgenfunction(obj)

        arguments: List[ArgumentDescriptor] = []
        is_pos_only_rendered = False
        is_kw_only_rendered = False
        is_self = callable_type in CLASS_FUNCTIONS
        for parameter in signature.parameters.values():
            if is_self:
                is_self = False
                continue

            if parameter.kind is Parameter.VAR_POSITIONAL:
                arg_name = '*' + parameter.name
                type_name = None
                default = ArgumentDescriptor.EMPTY
                docstring_param = None
            elif parameter.kind is Parameter.VAR_KEYWORD:
                arg_name = '**' + parameter.name
                type_name = None
                default = ArgumentDescriptor.EMPTY
                docstring_param = None
            else:
                if parameter.kind is Parameter.POSITIONAL_ONLY and not is_pos_only_rendered:
                    arguments.append(
                        ArgumentDescriptor(
                            CallableDescriptor.POSITIONAL_ONLY,
                            None,
                            None
                        )
                    )
                    is_pos_only_rendered = True
                elif parameter.kind is Parameter.KEYWORD_ONLY and not is_kw_only_rendered:
                    arguments.append(
                        ArgumentDescriptor(
                            CallableDescriptor.KEYWORD_ONLY,
                            None,
                            None
                        )
                    )
                    is_kw_only_rendered = True

                arg_name = parameter.name

                docstring_param = find_docstring_param(
                    parameter.name,
                    docstring
                )

                type_name = get_type_name(
                    parameter.annotation, docstring_param)

                if parameter.default is Parameter.empty:
                    default = ArgumentDescriptor.EMPTY
                elif prefer_docstring and docstring_param is not None:
                    default = docstring_param.default
                else:
                    default = parameter.default

            description = (
                docstring_param.description
                if docstring_param
                else None
            )

            arguments.append(
                ArgumentDescriptor(arg_name, type_name, description, default)
            )

        return_description: Optional[str] = None
        if callable_type == CallableType.CONSTRUCTOR or signature.return_annotation is None:
            return_type = 'None'
        elif not signature.return_annotation or signature.return_annotation == Parameter.empty:
            return_type = 'Any'
        else:
            return_type = get_type_name(
                signature.return_annotation,
                docstring.returns if docstring else None
            )
            return_description = (
                docstring.returns.description
                if docstring and docstring.returns
                else None
            )

        raises: Optional[List[RaisesDescriptor]] = [
            RaisesDescriptor(error.type_name, error.description)
            for error in docstring.raises
        ] if docstring and docstring.raises else None

        summary = docstring.short_description if docstring else None
        description = docstring.long_description if docstring else None

        examples: Optional[List[str]] = [
            meta.description
            for meta in docstring.meta
            if 'examples' in meta.args
        ] if docstring is not None else None

        module_obj = inspect.getmodule(obj)
        module = obj.__module__
        package = module_obj.__package__ if module_obj and module_obj.__package__ else None
        file = make_file_relative(
            module_obj.__file__
            if module_obj and hasattr(module_obj, '__file__')
            else None
        )

        if qualifier:
            name = obj.__name__
        elif hasattr(obj, '__qualname__'):
            qualifier, _, name = obj.__qualname__.rpartition('.')
            if not qualifier:
                qualifier = package
        else:
            qualifier = package
            name = obj.__name__

        return CallableDescriptor(
            qualifier or '',
            name,
            summary,
            description,
            arguments,
            return_type,
            return_description,
            callable_type,
            is_async,
            is_generator,
            raises,
            examples,
            module,
            package,
            file
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jetblack_markdown', 'jetblack_markdown.metadata']

package_data = \
{'': ['*'], 'jetblack_markdown': ['templates/*']}

install_requires = \
['docstring-parser>=0.15,<0.16', 'markdown>=3.3,<4.0']

setup_kwargs = {
    'name': 'jetblack-markdown',
    'version': '1.1.0',
    'description': 'A markdown extension for python documentation',
    'long_description': '# jetblack-markdown\n\nMarkdown extensions for automatic document generation\n(read the [docs](https://rob-blackbourn.github.io/jetblack-markdown/)).\n\n## Markdown Extension\n\nA markdown extension is provided for automatically documenting python code.\n\nModules are referred to as follows:\n\n```markdown\n# A Top Level Module\n\n@[jetblack_markdown]\n\n# A Package\n\n@[jetblack_markdown.autodoc]\n\n# A function\n\n@[jetblack_markdown.autodoc:makeExtension]\n\n# A class\n\n@[jetblack_markdown.autodoc.metadata:PropertyDescriptor]\n```\n\n## mkdocs integration\n\nThis site was generated using `mkdocs` and the following config:\n\n```yaml\nsite_name: jetblack-markdown\n\ndocs_dir: documentation\nsite_dir: docs\n\nmarkdown_extensions:\n  - admonition\n  - codehilite\n  - jetblack_markdown.autodoc:\n      class_from_init: true\n      ignore_dunder: true\n      ignore_private: true\n      ignore_all: false\n      prefer_docstring: true\n      template_folder: null\n\nextra_css:\n    - css/custom.css\n```\n\n### Configuration\n\nThere are some configuration parameters.\n\n* class_from_init (bool, optional): If True use the docstring from\n    the &#95;&#95;init&#95;&#95; function for classes. Defaults to\n    True.\n* ignore_dunder (bool, optional): If True ignore\n    &#95;&#95;XXX&#95;&#95; functions. Defaults to True.\n* ignore_private (bool, optional): If True ignore methods\n    (those prefixed &#95;XXX). Defaults to True.\n* ignore_all (bool): If True ignore the &#95;&#95;all&#95;&#95; member.\n* prefer_docstring (bool): If true prefer the docstring.\n* template_folder(Optional[str], optional): Specify a custom template folder.\n    The template "main.jinja2" will be rendered passing an `obj` parameter\n    which is a `jetblack.markdown.metadata.Descriptor`\n\n## Customizing\n\nAll the rendering is done with jinja2 templates. Start by copying the current\ntemplates from jetblack_markdown/templates and specify the `template_folder` in\nthe `mkdocs.yml`.\n\n',
    'author': 'Rob Blackbourn',
    'author_email': 'rob.blackbourn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rob-blackbourn/jetblack-markdown',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
